from enum import Enum
from typing import Annotated, Any, Literal, Optional, Tuple, Union

import numpy as np
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    HttpUrl,
    StringConstraints,
    TypeAdapter,
)
from pydantic.alias_generators import to_camel
from typing_extensions import TypeAlias

SplitIndicesType: TypeAlias = list[int]
"""
A split is defined by a sequence of integers.
"""

SplitType: TypeAlias = tuple[SplitIndicesType, Union[SplitIndicesType, dict[str, SplitIndicesType]]]
"""
A split is a pair of which the first item is always assumed to be the train set.
The second item can either be a single test set or a dictionary with multiple, named test sets.
"""

PredictionsType: TypeAlias = Union[np.ndarray, dict[str, Union[np.ndarray, dict[str, np.ndarray]]]]
"""
A prediction is one of three things:

- A single array (single-task, single test set)
- A dictionary of arrays (single-task, multiple test sets) 
- A dictionary of dictionaries of arrays (multi-task, multiple test sets)
"""

DatapointPartType = Union[Any, tuple[Any], dict[str, Any]]
DatapointType: TypeAlias = tuple[DatapointPartType, DatapointPartType]
"""
A datapoint has:

- A single input or multiple inputs (either as dict or tuple)
- No target, a single target or a multiple targets (either as dict or tuple)
"""

SlugStringType: TypeAlias = Annotated[
    str, StringConstraints(pattern="^[a-z0-9-]+$", min_length=4, max_length=64)
]
"""
A URL-compatible string that can serve as slug on the hub.
"""

SlugCompatibleStringType: TypeAlias = Annotated[
    str, StringConstraints(pattern="^[A-Za-z0-9_-]+$", min_length=4, max_length=64)
]
"""
A URL-compatible string that can be turned into a slug by the hub.

Can only use alpha-numeric characters, underscores and dashes. 
The string must be at least 4 and at most 64 characters long.
"""


HubUser: TypeAlias = SlugCompatibleStringType
"""
A user on the Polaris Hub is identified by a username, 
which is a [`SlugCompatibleStringType`][polaris.utils.types.SlugCompatibleStringType].
"""

HttpUrlAdapter = TypeAdapter(HttpUrl)
HttpUrlString: TypeAlias = Annotated[str, BeforeValidator(lambda v: HttpUrlAdapter.validate_python(v) and v)]

"""
A validated URL that will be turned into a string.
This is useful for interactions with httpx and authlib, who have their own URL types.
"""

DirectionType: TypeAlias = float | Literal["min", "max"]
"""
The direction of any variable to be sorted.
This can be used to sort the metric score, indicate the optmization direction of endpoint.
"""

AccessType: TypeAlias = Literal["public", "private"]
"""
Type to specify access to a dataset, benchmark or result in the Hub.
"""

TimeoutTypes = Union[Tuple[int, int], Literal["timeout", "never"]]
"""
Timeout types for specifying maximum wait times.
"""

IOMode: TypeAlias = Literal["r", "r+", "a", "w", "w-"]
"""
Type to specify the mode for input/output operations (I/O) when interacting with a file or resource.
"""

SupportedLicenseType: TypeAlias = Literal[
    "CC-BY-4.0", "CC-BY-SA-4.0", "CC-BY-NC-4.0", "CC-BY-NC-SA-4.0", "CC0-1.0", "MIT"
]
"""
Supported license types for dataset uploads to Polaris Hub
"""

ZarrConflictResolution: TypeAlias = Literal["raise", "replace", "skip"]
"""
Type to specify which action to take when encountering existing files within a Zarr archive.
"""

ChecksumStrategy: TypeAlias = Literal["verify", "verify_unless_zarr", "ignore"]
"""
Type to specify which action to take to verify the data integrity of an artifact through a checksum.
"""


class HubOwner(BaseModel):
    """An owner of an artifact on the Polaris Hub

    The slug is most important as it is the user-facing part of this data model.
    The externalId and type are added to be consistent with the model returned by the Polaris Hub .
    """

    slug: SlugStringType
    external_id: Optional[str] = None
    type: Optional[Literal["user", "organization"]] = None

    # Pydantic config
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    def __str__(self):
        return self.slug

    @staticmethod
    def normalize(owner: Union[str, "HubOwner"]) -> "HubOwner":
        """
        Normalize a string or `HubOwner` instance to a `HubOwner` instance.
        """
        return owner if isinstance(owner, HubOwner) else HubOwner(slug=owner)


class TargetType(Enum):
    """The high-level classification of different targets."""

    REGRESSION = "regression"
    CLASSIFICATION = "classification"


class TaskType(Enum):
    """The high-level classification of different tasks."""

    MULTI_TASK = "multi_task"
    SINGLE_TASK = "single_task"



from contextlib import contextmanager
from yaspin import yaspin
from yaspin.spinners import Spinners


@contextmanager
def tmp_attribute_change(obj, attribute, value):
    """Temporarily set and reset an attribute of an object."""
    original_value = getattr(obj, attribute)
    try:
        setattr(obj, attribute, value)
        yield obj
    finally:
        setattr(obj, attribute, original_value)


class ProgressIndicator:
    def __init__(self, success_msg: str, error_msg: str, start_msg: str = "In progress..."):
        self._start_msg = start_msg
        self._success_msg = success_msg
        self._error_msg = error_msg

        self._spinner = yaspin(Spinners.dots, text=self._start_msg)

    def __enter__(self):
        self._spinner.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._spinner.text = ""

        if exc_type:
            self._spinner.red.fail(f"💥 ERROR: {self._error_msg}")
        else:
            self._spinner.green.ok(f"✅ SUCCESS: \033[1m{self._success_msg}\033[0m\n")

        self._spinner.stop()

    def update_success_msg(self, msg: str):
        self._success_msg = msg



import datamol as dm
import numpy as np
import pytest
import zarr
from datamol.utils import fs

import polaris as po
from polaris.benchmark import (
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)
from polaris.dataset import ColumnAnnotation, Dataset
from polaris.utils.types import HubOwner


def check_version(artifact):
    assert po.__version__ == artifact.polaris_version


@pytest.fixture(scope="module")
def test_data():
    data = dm.data.freesolv()[:100]
    # set an abitrary threshold for testing purpose.
    data["CLASS_expt"] = data["expt"].gt(0).astype(int).values
    data["CLASS_calc"] = data["calc"].gt(0).astype(int).values
    data["MULTICLASS_expt"] = np.random.randint(low=0, high=3, size=data.shape[0])
    data["MULTICLASS_calc"] = np.random.randint(low=0, high=3, size=data.shape[0])
    return data


@pytest.fixture(scope="module")
def caffeine():
    # Let's generate a toy dataset with a single molecule
    smiles = "Cn1cnc2c1c(=O)n(C)c(=O)n2C"
    mol = dm.to_mol(smiles)

    # We will generate 3D conformers for this molecule with some conformers
    # NOTE (cwognum): We only generate a single conformer, because dm.to_sdf() only saves one.
    mol = dm.conformers.generate(mol, align_conformers=True, n_confs=1)

    # Let's also set a molecular property
    mol.SetProp("my_property", "my_value")
    return mol


@pytest.fixture(scope="module")
def sdf_file(tmp_path_factory, caffeine):
    path = tmp_path_factory.mktemp("data") / "caffeine.sdf"
    dm.to_sdf(caffeine, path)
    return path


@pytest.fixture(scope="module")
def test_org_owner():
    return HubOwner(organizationId="test-organization", slug="test-organization")


@pytest.fixture(scope="module")
def test_user_owner():
    return HubOwner(userId="test-user", slug="test-user")


@pytest.fixture(scope="function")
def test_dataset(test_data, test_org_owner):
    dataset = Dataset(
        table=test_data,
        name="test-dataset",
        source="https://www.example.com",
        annotations={"expt": ColumnAnnotation(user_attributes={"unit": "kcal/mol"})},
        tags=["tagA", "tagB"],
        user_attributes={"attributeA": "valueA", "attributeB": "valueB"},
        owner=test_org_owner,
        license="CC-BY-4.0",
        curation_reference="https://www.example.com",
    )
    check_version(dataset)
    return dataset


@pytest.fixture(scope="function")
def zarr_archive(tmp_path):
    tmp_path = fs.join(tmp_path, "data.zarr")
    root = zarr.open(tmp_path, mode="w")
    root.array("A", data=np.random.random((100, 2048)), chunks=(1, None))
    root.array("B", data=np.random.random((100, 2048)), chunks=(1, None))
    zarr.consolidate_metadata(root.store)
    return tmp_path


@pytest.fixture(scope="function")
def test_single_task_benchmark(test_dataset):
    train_indices = list(range(90))
    test_indices = list(range(90, 100))
    benchmark = SingleTaskBenchmarkSpecification(
        name="single-task-benchmark",
        dataset=test_dataset,
        metrics=[
            "mean_absolute_error",
            "mean_squared_error",
            "r2",
            "spearmanr",
            "pearsonr",
            "explained_var",
            "absolute_average_fold_error",
        ],
        main_metric="mean_absolute_error",
        split=(train_indices, test_indices),
        target_cols="expt",
        input_cols="smiles",
    )
    check_version(benchmark)
    return benchmark


@pytest.fixture(scope="function")
def test_single_task_benchmark_clf(test_dataset):
    train_indices = list(range(90))
    test_indices = list(range(90, 100))
    benchmark = SingleTaskBenchmarkSpecification(
        name="single-task-benchmark",
        dataset=test_dataset,
        main_metric="accuracy",
        metrics=["accuracy", "f1", "roc_auc", "pr_auc", "mcc", "cohen_kappa", "balanced_accuracy"],
        split=(train_indices, test_indices),
        target_cols="CLASS_expt",
        input_cols="smiles",
    )
    check_version(benchmark)
    return benchmark


@pytest.fixture(scope="function")
def test_single_task_benchmark_multi_clf(test_dataset):
    np.random.seed(111)
    indices = np.arange(100)
    np.random.shuffle(indices)
    train_indices = indices[:80]
    test_indices = indices[80:]

    benchmark = SingleTaskBenchmarkSpecification(
        name="single-task-benchmark",
        dataset=test_dataset,
        main_metric="accuracy",
        metrics=[
            "accuracy",
            "balanced_accuracy",
            "mcc",
            "cohen_kappa",
            "f1_macro",
            "f1_micro",
            "roc_auc_ovr",
            "roc_auc_ovo",
            "pr_auc",
        ],
        split=(train_indices, test_indices),
        target_cols="MULTICLASS_expt",
        input_cols="smiles",
    )
    check_version(benchmark)
    return benchmark


@pytest.fixture(scope="function")
def test_single_task_benchmark_multiple_test_sets(test_dataset):
    train_indices = list(range(90))
    test_indices = {"test_1": list(range(90, 95)), "test_2": list(range(95, 100))}
    benchmark = SingleTaskBenchmarkSpecification(
        name="single-task-benchmark",
        dataset=test_dataset,
        metrics=[
            "mean_absolute_error",
            "mean_squared_error",
            "r2",
            "spearmanr",
            "pearsonr",
            "explained_var",
            "absolute_average_fold_error",
        ],
        main_metric="r2",
        split=(train_indices, test_indices),
        target_cols="expt",
        input_cols="smiles",
    )
    check_version(benchmark)
    return benchmark


@pytest.fixture(scope="function")
def test_single_task_benchmark_clf_multiple_test_sets(test_dataset):
    np.random.seed(111)  # make sure two classes in `y_true`
    indices = np.arange(100)
    np.random.shuffle(indices)
    train_indices = indices[:80]
    test_indices = {"test_1": indices[80:90], "test_2": indices[90:]}
    benchmark = SingleTaskBenchmarkSpecification(
        name="single-task-benchmark-clf",
        dataset=test_dataset,
        metrics=["accuracy", "f1", "roc_auc", "pr_auc", "mcc", "cohen_kappa"],
        main_metric="pr_auc",
        split=(train_indices, test_indices),
        target_cols="CLASS_calc",
        input_cols="smiles",
    )
    check_version(benchmark)
    return benchmark


@pytest.fixture(scope="function")
def test_multi_task_benchmark(test_dataset):
    # For the sake of simplicity, just use a small set of indices
    train_indices = list(range(90))
    test_indices = list(range(90, 100))
    benchmark = MultiTaskBenchmarkSpecification(
        name="multi-task-benchmark",
        dataset=test_dataset,
        main_metric="mean_absolute_error",
        metrics=[
            "mean_absolute_error",
            "mean_squared_error",
            "r2",
            "spearmanr",
            "pearsonr",
            "explained_var",
            "absolute_average_fold_error",
        ],
        split=(train_indices, test_indices),
        target_cols=["expt", "calc"],
        input_cols="smiles",
        target_types={"expt": "regression"},
    )
    check_version(benchmark)
    return benchmark


@pytest.fixture(scope="function")
def test_multi_task_benchmark_clf(test_dataset):
    # For the sake of simplicity, just use a small set of indices
    train_indices = list(range(90))
    test_indices = list(range(90, 100))
    benchmark = MultiTaskBenchmarkSpecification(
        name="multi-task-benchmark",
        dataset=test_dataset,
        main_metric="accuracy",
        metrics=["accuracy", "f1", "roc_auc", "pr_auc", "mcc", "cohen_kappa"],
        split=(train_indices, test_indices),
        target_cols=["CLASS_expt", "CLASS_calc"],
        input_cols="smiles",
    )
    check_version(benchmark)
    return benchmark



import os
import sys

from loguru import logger

from ._version import __version__
from .loader import load_benchmark, load_dataset

__all__ = ["load_dataset", "load_benchmark", "__version__"]

# Configure the default logging level
os.environ["LOGURU_LEVEL"] = os.environ.get("LOGURU_LEVEL", "INFO")
logger.remove()
logger.add(sys.stderr, level=os.environ["LOGURU_LEVEL"])



from polaris.dataset.converters._base import Converter
from polaris.dataset.converters._sdf import SDFConverter
from polaris.dataset.converters._zarr import ZarrConverter

__all__ = ["Converter", "SDFConverter", "ZarrConverter"]



from typing import Annotated

import typer

from polaris.hub.client import PolarisHubClient
from polaris.hub.settings import PolarisHubSettings

app = typer.Typer(
    add_completion=False,
    help="Polaris is a framework for benchmarking methods in drug discovery.",
)


@app.command("login")
def login(
    client_env_file: Annotated[
        str, typer.Option(help="Environment file to overwrite the default environment variables")
    ] = ".env",
    auto_open_browser: Annotated[
        bool, typer.Option(help="Whether to automatically open the link in a browser to retrieve the token")
    ] = True,
    overwrite: Annotated[
        bool, typer.Option(help="Whether to overwrite the access token if you are already logged in")
    ] = False,
):
    """Authenticate to the Polaris Hub.

    This CLI will use the OAuth2 protocol to gain token-based access to the Polaris Hub API.
    """
    with PolarisHubClient(settings=PolarisHubSettings(_env_file=client_env_file)) as client:
        client.login(auto_open_browser=auto_open_browser, overwrite=overwrite)


@app.command(hidden=True)
def secret():
    # NOTE (cwognum): Empty, hidden command to force Typer to not collapse the subcommand.
    # Added because I anticipate we will want to add more subcommands later on. This will keep
    # the API consistent in the meantime. Once there are other subcommands, it can be removed.
    # See also: https://github.com/tiangolo/typer/issues/315
    raise NotImplementedError()


if __name__ == "__main__":
    app()



import json
from typing import Dict, Optional, Union

import fsspec
from loguru import logger
from packaging.version import Version
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_serializer,
    field_validator,
)
from pydantic.alias_generators import to_camel

import polaris as po
from polaris.utils.misc import sluggify
from polaris.utils.types import HubOwner, SlugCompatibleStringType


class BaseArtifactModel(BaseModel):
    """
    Base class for all artifacts on the Hub. Specifies meta-data that is used by the Hub.

    Info: Optional
        Despite all artifacts basing this class, note that all attributes are optional.
        This ensures the library can be used without the Polaris Hub.
        Only when uploading to the Hub, some of the attributes are required.

    Attributes:
        name: A slug-compatible name for the dataset.
            Together with the owner, this is used by the Hub to uniquely identify the benchmark.
        description: A beginner-friendly, short description of the dataset.
        tags: A list of tags to categorize the benchmark by. This is used by the hub to search over benchmarks.
        user_attributes: A dict with additional, textual user attributes.
        owner: A slug-compatible name for the owner of the dataset.
            If the dataset comes from the Polaris Hub, this is the associated owner (organization or user).
            Together with the name, this is used by the Hub to uniquely identify the benchmark.
        polaris_version: The version of the Polaris library that was used to create the artifact.
    """

    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True, arbitrary_types_allowed=True)

    name: Optional[SlugCompatibleStringType] = None
    description: str = ""
    tags: list[str] = Field(default_factory=list)
    user_attributes: Dict[str, str] = Field(default_factory=dict)
    owner: Optional[HubOwner] = None
    polaris_version: str = po.__version__

    @computed_field
    @property
    def artifact_id(self) -> Optional[str]:
        return f"{self.owner}/{sluggify(self.name)}" if self.owner and self.name else None

    @field_validator("polaris_version")
    @classmethod
    def _validate_version(cls, value: str) -> str:
        if value != "dev":
            # Make sure it is a valid semantic version
            Version(value)

        current_version = po.__version__
        if value != current_version:
            logger.info(
                f"The version of Polaris that was used to create the artifact ({value}) is different "
                f"from the currently installed version of Polaris ({current_version})."
            )
        return value

    @field_validator("owner", mode="before")
    @classmethod
    def _validate_owner(cls, value: Union[str, HubOwner, None]):
        if isinstance(value, str):
            return HubOwner(slug=value)
        return value

    @field_serializer("owner")
    def _serialize_owner(self, value: HubOwner) -> Union[str, None]:
        return value.slug if value else None

    @classmethod
    def from_json(cls, path: str):
        """Loads a benchmark from a JSON file.

        Args:
            path: Loads a benchmark specification from a JSON file.
        """
        with fsspec.open(path, "r") as f:
            data = json.load(f)
        return cls.model_validate(data)

    def to_json(self, path: str):
        """Saves the benchmark to a JSON file.

        Args:
            path: Saves the benchmark specification to a JSON file.
        """
        with fsspec.open(path, "w") as f:
            json.dump(self.model_dump(), f)



"""
JSON 2 HTML Converter

(c) Varun Malhotra 2013-2021
Source Code: https://github.com/softvar/json2html

Contributors:
1. Michel Müller (@muellermichel), https://github.com/softvar/json2html/pull/2
2. Daniel Lekic (@lekic), https://github.com/softvar/json2html/pull/17

LICENSE: MIT

- 30/06/2023: adapted from: https://github.com/softvar/json2html/blob/e1feea273b210d11e4b2e59b9778a0bb4845fbd4/json2html/jsonconv.py
"""

from html import escape as html_escape

text = str
text_types = (str,)


class Dict2Html:
    def convert(
        self,
        data_dict: dict,
        table_attributes='border="1"',
        clubbing=True,
        encode=False,
        escape=True,
    ):
        """
        Convert JSON to HTML Table format
        """
        # table attributes such as class, id, data-attr-*, etc.
        # eg: table_attributes = 'class = "table table-bordered sortable"'
        self.table_init_markup = "<table %s>" % table_attributes
        self.clubbing = clubbing
        self.escape = escape
        json_input = data_dict
        converted = self.convert_json_node(json_input)
        if encode:
            return converted.encode("ascii", "xmlcharrefreplace")
        return converted

    def column_headers_from_list_of_dicts(self, json_input):
        """
        This method is required to implement clubbing.
        It tries to come up with column headers for your input
        """
        if not json_input or not hasattr(json_input, "__getitem__") or not hasattr(json_input[0], "keys"):
            return None
        column_headers = json_input[0].keys()
        for entry in json_input:
            if (
                not hasattr(entry, "keys")
                or not hasattr(entry, "__iter__")
                or len(entry.keys()) != len(column_headers)
            ):
                return None
            for header in column_headers:
                if header not in entry:
                    return None
        return column_headers

    def convert_json_node(self, json_input):
        """
        Dispatch JSON input according to the outermost type and process it
        to generate the super awesome HTML format.
        We try to adhere to duck typing such that users can just pass all kinds
        of funky objects to json2html that *behave* like dicts and lists and other
        basic JSON types.
        """
        if type(json_input) in text_types:
            if self.escape:
                return html_escape(text(json_input))
            else:
                return text(json_input)
        if hasattr(json_input, "items"):
            return self.convert_object(json_input)
        if hasattr(json_input, "__iter__") and hasattr(json_input, "__getitem__"):
            return self.convert_list(json_input)
        return text(json_input)

    def convert_list(self, list_input):
        """
        Iterate over the JSON list and process it
        to generate either an HTML table or a HTML list, depending on what's inside.
        If suppose some key has array of objects and all the keys are same,
        instead of creating a new row for each such entry,
        club such values, thus it makes more sense and more readable table.

        @example:
            jsonObject = {
                "sampleData": [
                    {"a":1, "b":2, "c":3},
                    {"a":5, "b":6, "c":7}
                ]
            }
            OUTPUT:
            _____________________________
            |               |   |   |   |
            |               | a | c | b |
            |   sampleData  |---|---|---|
            |               | 1 | 3 | 2 |
            |               | 5 | 7 | 6 |
            -----------------------------

        @contributed by: @muellermichel
        """
        if not list_input:
            return ""
        converted_output = ""
        column_headers = None
        if self.clubbing:
            column_headers = self.column_headers_from_list_of_dicts(list_input)
        if column_headers is not None:
            converted_output += self.table_init_markup
            converted_output += "<thead>"
            converted_output += "<tr><th>" + "</th><th>".join(column_headers) + "</th></tr>"
            converted_output += "</thead>"
            converted_output += "<tbody>"
            for list_entry in list_input:
                converted_output += "<tr><td>"
                converted_output += "</td><td>".join(
                    [self.convert_json_node(list_entry[column_header]) for column_header in column_headers]
                )
                converted_output += "</td></tr>"
            converted_output += "</tbody>"
            converted_output += "</table>"
            return converted_output

        # so you don't want or need clubbing eh? This makes @muellermichel very sad... ;(
        # alright, let's fall back to a basic list here...
        converted_output = "<ul><li>"
        converted_output += "</li><li>".join([self.convert_json_node(child) for child in list_input])
        converted_output += "</li></ul>"
        return converted_output

    def convert_object(self, json_input):
        """
        Iterate over the JSON object and process it
        to generate the super awesome HTML Table format
        """
        if not json_input:
            return ""  # avoid empty tables
        converted_output = self.table_init_markup + "<tr>"
        converted_output += "</tr><tr>".join(
            [
                "<th>%s</th><td>%s</td>" % (self.convert_json_node(k), self.convert_json_node(v))
                for k, v in json_input.items()
            ]
        )
        converted_output += "</tr></table>"
        return converted_output


dict2html = Dict2Html().convert



from polaris.dataset._column import ColumnAnnotation, Modality
from polaris.dataset._dataset import Dataset
from polaris.dataset._factory import DatasetFactory, create_dataset_from_file
from polaris.dataset._subset import Subset

__all__ = [
    "ColumnAnnotation",
    "Dataset",
    "Subset",
    "Modality",
    "DatasetFactory",
    "create_dataset_from_file",
]



from enum import Enum, auto, unique

import datamol as dm

# Map of conversion operations which can be applied to dataset columns
conversion_map = {"SMILES_TO_MOL": dm.to_mol, "BYTES_TO_MOL": dm.Mol}


@unique
class Adapter(Enum):
    """
    Adapters are predefined callables that change the format of the data.
    Adapters are serializable and can thus be saved alongside datasets.

    Attributes:
        SMILES_TO_MOL: Convert a SMILES string to a RDKit molecule.
        BYTES_TO_MOL: Convert a RDKit binary string to a RDKit molecule.
    """

    SMILES_TO_MOL = auto()
    BYTES_TO_MOL = auto()

    def __call__(self, data):
        if isinstance(data, tuple):
            return tuple(conversion_map[self.name](d) for d in data)
        return conversion_map[self.name](data)



"""
The code in this file is based on the zarr-checksum package

Mainted by Jacob Nesbitt, released under the DANDI org on Github
and with Kitware, Inc. credited as the author. This code is released
with the Apache 2.0 license.

See also: https://github.com/dandi/zarr_checksum

Instead of adding the package as a dependency, we opted to copy over the code
because it is a small and self-contained module that we will want to alter to
support our Polaris code base.

NOTE: We have made some modifications to the original code.

----

Copyright 2023 Kitware, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import hashlib
import heapq
import os
import re
from dataclasses import asdict, dataclass, field
from functools import total_ordering
from json import dumps
from pathlib import Path
from typing import List, Tuple

import fsspec
import zarr
import zarr.errors
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel
from tqdm import tqdm

from polaris.utils.errors import InvalidZarrChecksum

ZARR_DIGEST_PATTERN = "([0-9a-f]{32})-([0-9]+)-([0-9]+)"


def compute_zarr_checksum(zarr_root_path: str) -> Tuple[str, List["ZarrFileChecksum"]]:
    r"""
    Implements an algorithm to compute the Zarr checksum.

    Warning: This checksum is sensitive to Zarr configuration. 
        This checksum is sensitive to change in the Zarr structure. For example, if you change the chunk size, 
        the checksum will also change.

    To understand how this works, consider the following directory structure:

               . (root)
              / \
             a   c
            /
           b
    
    Within zarr, this would for example be:

    - `root`: A Zarr Group with a single Array.
    - `a`: A Zarr Array
    - `b`: A single chunk of the Zarr Array
    - `c`: A metadata file (i.e. .zarray, .zattrs or .zgroup) 

    To compute the checksum, we first find all the trees in the node, in this case b and c. 
    We compute the hash of the content (the raw bytes) for each of these files.
    
    We then work our way up the tree. For any node (directory), we find all children of that node.
    In an sorted order, we then serialize a list with - for each of the children - the checksum, size, and number of children.
    The hash of the directory is then equal to the hash of the serialized JSON.

    The Polaris implementation is heavily based on the [`zarr-checksum` package](https://github.com/dandi/zarr_checksum).
    This method is the biggest deviation of the original code.
    """

    # Get the protocol of the path
    protocol = fsspec.utils.get_protocol(zarr_root_path)

    # We only support computing checksum for local datasets.
    # NOTE (cwognum): We don't have to verify the checksum for datasets stored on the Hub,
    # as the Hub will do this on upload. And if you're streaming the data from the Hub,
    # we will check the checksum of each chunk on download.
    if protocol != "file":
        raise RuntimeError(
            "You can only compute the checksum for a local Zarr archive. "
            "You can cache a dataset to your local machine with `dataset.cache()`."
        )

    # Normalize the path
    zarr_root_path = os.path.expandvars(zarr_root_path)
    zarr_root_path = os.path.expanduser(zarr_root_path)
    zarr_root_path = os.path.abspath(zarr_root_path)

    fs, zarr_root_path = fsspec.url_to_fs(zarr_root_path)

    # Make sure the path exists and is a Zarr archive
    zarr.open_group(zarr_root_path, mode="r")

    # Generate the checksum
    tree = _ZarrChecksumTree()

    # Find all files below the root
    leaves = fs.find(zarr_root_path, detail=True)
    zarr_md5sum_manifest = []

    for file in tqdm(leaves.values(), desc="Finding all files in the Zarr archive"):
        path = file["name"]

        relpath = path.removeprefix(zarr_root_path)
        relpath = relpath.lstrip("/")
        relpath = Path(relpath)

        size = file["size"]

        # Compute md5sum of file
        md5sum = hashlib.md5()
        with fs.open(path, "rb") as f:
            for chunk in iter(lambda: f.read(8192), b""):
                md5sum.update(chunk)
        digest = md5sum.hexdigest()

        # Add a leaf to the tree
        # (This actually adds the file's checksum to the parent directory's manifest)
        tree.add_leaf(
            path=relpath,
            size=size,
            digest=digest,
        )

        # We persist the checksums for leaf nodes separately,
        # because this is what the Hub needs to verify data integrity.
        zarr_md5sum_manifest.append(ZarrFileChecksum(path=str(relpath), md5sum=digest, size=size))

    # Compute digest
    return tree.process().digest, zarr_md5sum_manifest


class ZarrFileChecksum(BaseModel):
    """
    This data is sent to the Hub to verify the integrity of the Zarr archive on upload.

    Attributes:
        path: The path of the file relative to the Zarr root.
        md5sum: The md5sum of the file.
        size: The size of the file in bytes.
    """

    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True, arbitrary_types_allowed=True)

    path: str
    md5sum: str
    size: int


# ================================
# Overview of the data structures
# ================================

# NOTE (cwognum): I kept forgetting how this works, so I'm writing it down
# - The ZarrChecksumTree is a binary tree (heap queue). It determines the order in which to process the nodes.
# - The ZarrChecksumNode is a node in the ZarrChecksumTree queue. It represents a directory in the Zarr archive and
#   stores a manifest with all the data needed to compute the checksum for that node.
# - The ZarrChecksumManifest is a collection of checksums for all direct (non-recursive) children of a directory.
# - The ZarrChecksum is the data used to compute the checksum for a file or directory in a Zarr Archive.
#   This is the object that the ZarrChecksumManifest stores a collection of.
# - A ZarrDirectoryDigest is the result of processing a directory. Once completed,
#   it is added to the ZarrChecksumManifest of its parent as part of a ZarrChecksum.

# NOTE (cwognum): As a first impression, it seems there is some redundancy in the data structures.
#   My feeling is that we could reduce the redundancy to simplify things and improve maintainability.
#   However, for the time being, let's stick close to the original code.

# ================================


# Pydantic models aren't used for performance reasons
class _ZarrChecksumTree:
    """
    The ZarrChecksumTree is a tree structure that maintains the state of the checksum algorithm.

    Initialized with a set of leafs (i.e. files), the nodes in this tree correspond to all directories
    that are above those leafs and below the Zarr Root.

    The tree then implements the logic for retrieving the next node (i.e. directory) to process,
    and for computing the checksum for that node based on its children.
    Once it reaches the root, it has computed the checksum for the entire Zarr archive.
    """

    def __init__(self) -> None:
        # Queue to prioritize the next node to process
        self._heap: list[tuple[int, _ZarrChecksumNode]] = []

        # Map of (relative) paths to nodes.
        self._path_map: dict[Path, _ZarrChecksumNode] = {}

    @property
    def empty(self) -> bool:
        """Check if the tree is empty."""
        # This is used as an exit condition in the process() method
        return len(self._heap) == 0

    def _add_path(self, key: Path) -> None:
        """Adds a new entry to the heap queue for which we need to compute the checksum."""

        # Create a new node
        # A node represents a file or directory.
        # A node refers to a node in the heap queue (i.e. binary tree)
        # The structure of the heap is thus _not_ the same as the structure of the file system!
        node = _ZarrChecksumNode(path=key, checksums=_ZarrChecksumManifest())
        self._path_map[key] = node

        # Add node to heap with length (negated to represent a max heap)
        # We use the length of the parents (relative to the Zarr root) to structure the heap.
        # The node with the longest path is the deepest node in the tree.
        # This node will be prioritized for processing next.
        length = len(key.parents)
        heapq.heappush(self._heap, (-1 * length, node))

    def _get_path(self, key: Path) -> "_ZarrChecksumNode":
        """
        If an entry for this path already exists, return it.
        Otherwise create a new one and return that.
        """
        if key not in self._path_map:
            self._add_path(key)
        return self._path_map[key]

    def add_leaf(self, path: Path, size: int, digest: str) -> None:
        """Add a leaf file to the tree."""
        parent_node = self._get_path(path.parent)
        parent_node.checksums.files.append(_ZarrChecksum(name=path.name, size=size, digest=digest))

    def add_node(self, path: Path, size: int, digest: str, count: int) -> None:
        """Add an internal node to the tree."""
        parent_node = self._get_path(path.parent)
        parent_node.checksums.directories.append(
            _ZarrChecksum(
                name=path.name,
                size=size,
                digest=digest,
                count=count,
            )
        )

    def pop_deepest(self) -> "_ZarrChecksumNode":
        """
        Returns the node with the highest priority for processing next.

        Returns (one of the) node(s) with the most parent directories
        (i.e. the deepest directory in the file system)
        """
        _, node = heapq.heappop(self._heap)
        del self._path_map[node.path]
        return node

    def process(self) -> "_ZarrDirectoryDigest":
        """Process the tree, returning the resulting top level digest."""

        # Begin with empty root node, so that if no files are present, the empty checksum is returned
        node = _ZarrChecksumNode(path=Path("."), checksums=_ZarrChecksumManifest())

        while not self.empty:
            # Get the next directory to process
            # Priority is based on the number of parents a directory has
            # In other word, the depth of the directory in the file system.
            node = self.pop_deepest()

            # If we have reached the root node, then we're done.
            if node.path == Path(".") or node.path == Path("/"):
                break

            # Add the parent of this node to the tree
            directory_digest = node.checksums.generate_digest()
            self.add_node(
                path=node.path,
                size=directory_digest.size,
                digest=directory_digest.digest,
                count=directory_digest.count,
            )

        # Return digest
        return node.checksums.generate_digest()


@dataclass
class _ZarrChecksumNode:
    """
    A node in the ZarrChecksumTree.

    This node represents a file or directory in the Zarr archive,
    but "node" here refers to a node in the heap queue (i.e. binary tree).
    The structure of the heap is thus _not_ the same as the structure of the file system!

    The node stores a manifest of checksums for all files and directories below it.
    """

    path: Path
    checksums: "_ZarrChecksumManifest"

    def __lt__(self, other: "_ZarrChecksumNode") -> bool:
        return str(self.path) < str(other.path)


@dataclass
class _ZarrChecksumManifest:
    """
    For a directory in the Zarr archive (i.e. a node in the heap queue),
    we maintain a manifest of the checksums for all files and directories
    below that directory.

    This data is then used to calculate the checksum of a directory.
    """

    directories: list["_ZarrChecksum"] = field(default_factory=list)
    files: list["_ZarrChecksum"] = field(default_factory=list)

    @property
    def is_empty(self) -> bool:
        return not (self.files or self.directories)

    def generate_digest(self) -> "_ZarrDirectoryDigest":
        """Generate an aggregated digest for the provided files/directories."""

        # Sort everything to ensure the checksum is deterministic
        self.files.sort()
        self.directories.sort()

        # Aggregate total file count
        count = len(self.files) + sum(checksum.count for checksum in self.directories)

        # Aggregate total size
        size = sum(file.size for file in self.files) + sum(directory.size for directory in self.directories)

        # Serialize json without any spacing
        json = dumps(asdict(self), separators=(",", ":"))

        # Generate digest
        md5 = hashlib.md5(json.encode("utf-8")).hexdigest()

        # Construct and return
        return _ZarrDirectoryDigest(md5=md5, count=count, size=size)


@total_ordering
@dataclass
class _ZarrChecksum:
    """
    The data used to compute the checksum for a file or directory in a Zarr Archive.

    This class is serialized to JSON, and as such, key order should not be modified.
    """

    digest: str
    name: str
    size: int
    count: int = 0

    # To make this class sortable
    def __lt__(self, other: "_ZarrChecksum") -> bool:
        return self.name < other.name


@dataclass
class _ZarrDirectoryDigest:
    """
    The digest for a directory in a Zarr Archive.

    The digest is a string representation that serves as a checksum for the directory.
    This is a utility class to (de)serialize that string.
    """

    md5: str
    count: int
    size: int

    @classmethod
    def parse(cls, checksum: str | None) -> "_ZarrDirectoryDigest":
        if checksum is None:
            return cls.parse(EMPTY_CHECKSUM)

        match = re.match(ZARR_DIGEST_PATTERN, checksum)
        if match is None:
            raise InvalidZarrChecksum()

        md5, count, size = match.groups()
        return cls(md5=md5, count=int(count), size=int(size))

    def __str__(self) -> str:
        return self.digest

    @property
    def digest(self) -> str:
        return f"{self.md5}-{self.count}-{self.size}"


# The "null" zarr checksum
EMPTY_CHECKSUM = _ZarrChecksumManifest().generate_digest().digest



import uuid
from typing import TYPE_CHECKING, Optional, Sequence

import datamol as dm
import pandas as pd
from rdkit import Chem

from polaris.dataset import ColumnAnnotation, Modality
from polaris.dataset._adapters import Adapter
from polaris.dataset.converters._base import Converter, FactoryProduct

if TYPE_CHECKING:
    from polaris.dataset import DatasetFactory


class SDFConverter(Converter):
    """
    Converts a SDF file into a Polaris dataset.

    Info: Binary strings for serialization
        This class converts the molecules to binary strings (for ML purposes, this should be lossless).
        This might not be the most storage efficient, but is fastest and easiest to maintain.
        See this [Github Discussion](https://github.com/rdkit/rdkit/discussions/7235) for more info.

    Properties defined on the molecule level in the SDF file can be extracted into separate columns
    or can be kept in the molecule object.

    Args:
        mol_column: The name of the column that will contain the pointers to the molecules.
        smiles_column: The name of the column that will contain the SMILES strings.
        use_isomeric_smiles: Whether to use isomeric SMILES.
        mol_id_column: The name of the column that will contain the molecule names.
        mol_prop_as_cols: Whether to extract properties defined on the molecule level in the SDF file into separate columns.
        groupby_key: The name of the column to group by. If set, the dataset can combine multiple pointers
            to the molecules into a single datapoint.
    """

    def __init__(
        self,
        mol_column: str = "molecule",
        smiles_column: Optional[str] = "smiles",
        use_isomeric_smiles: bool = True,
        mol_id_column: Optional[str] = None,
        mol_prop_as_cols: bool = True,
        groupby_key: Optional[str] = None,
        n_jobs: int = 1,
        zarr_chunks: Sequence[Optional[int]] = (1,),
    ) -> None:
        super().__init__()
        self.mol_column = mol_column
        self.smiles_column = smiles_column
        self.use_isomeric_smiles = use_isomeric_smiles
        self.mol_id_column = mol_id_column
        self.mol_prop_as_cols = mol_prop_as_cols
        self.groupby_key = groupby_key
        self.n_jobs = n_jobs
        self.chunks = zarr_chunks

    def convert(self, path: str, factory: "DatasetFactory") -> FactoryProduct:
        tmp_col = uuid.uuid4().hex

        # We do not sanitize the molecules or remove the Hs.
        # We assume the SDF has been processed by the user already and do not want to change it.
        df = dm.read_sdf(
            path,
            as_df=self.mol_prop_as_cols,
            smiles_column=self.smiles_column,
            mol_column=tmp_col,
            remove_hs=False,
            sanitize=False,
        )

        if not isinstance(df, pd.DataFrame):
            df = pd.DataFrame({tmp_col: df})

        if self.mol_column in df.columns:
            raise ValueError(
                f"The column name '{self.mol_column}' clashes with the name of a property in the SDF file. "
                f"Please choose another name by setting the `mol_column` in the {self.__class__.__name__}."
            )

        # Add a column with the molecule name if it doesn't exist yet
        if self.mol_id_column is not None and self.mol_id_column not in df.columns:

            def _get_name(mol: dm.Mol):
                return mol.GetProp(self.mol_id_column) if mol.HasProp(self.mol_id_column) else None

            names = dm.parallelized(_get_name, df[tmp_col], n_jobs=self.n_jobs, scheduler="threads")
            df[self.mol_id_column] = names

        # Add a column with the SMILES if it doesn't exist yet
        if self.smiles_column is not None and self.smiles_column not in df.columns:
            names = dm.parallelized(
                lambda mol: dm.to_smiles(mol, isomeric=self.use_isomeric_smiles),
                df[tmp_col],
                n_jobs=self.n_jobs,
            )
            df[self.smiles_column] = names

        # Convert the molecules to binary strings. This should be lossless and efficient.
        # NOTE (cwognum): We might want to not always store private and computed properties.
        props = Chem.PropertyPickleOptions.AllProps
        if self.mol_prop_as_cols:
            props &= ~Chem.PropertyPickleOptions.MolProps
        bytes_data = [mol.ToBinary(props) for mol in df[tmp_col]]

        df.drop(columns=[tmp_col], inplace=True)

        # Create the zarr array
        factory.zarr_root.array(self.mol_column, bytes_data, dtype=bytes, chunks=self.chunks)

        # Add a pointer column to the table
        # We support grouping by a key, to allow inputs of variable length
        grouped = pd.DataFrame(columns=[*df.columns, self.mol_column])
        if self.groupby_key is not None:
            for _, group in df.reset_index(drop=True).groupby(by=self.groupby_key):
                start = group.index[0]
                end = group.index[-1]

                if group.nunique().sum() != len(group.columns):
                    raise ValueError(
                        f"After grouping by {self.groupby_key}, values for other columns are not unique within a group. "
                        "Please handle this manually to ensure aggregation is done correctly."
                    )

                # Get the pointer path
                pointer_idx = f"{start}:{end}" if start != end else f"{start}"
                pointer = self.get_pointer(self.mol_column, pointer_idx)

                # Get the single unique value per column for the group and append
                unique_values = [group[col].unique()[0] for col in df.columns]
                grouped.loc[len(grouped)] = [*unique_values, pointer]

            df = grouped

        else:
            pointers = [self.get_pointer(self.mol_column, i) for i in range(len(df))]
            df[self.mol_column] = pd.Series(pointers)

        # Set the annotations
        annotations = {self.mol_column: ColumnAnnotation(is_pointer=True, modality=Modality.MOLECULE_3D)}
        if self.smiles_column is not None:
            annotations[self.smiles_column] = ColumnAnnotation(modality=Modality.MOLECULE)

        # Return the dataframe and the annotations
        return df, annotations, {self.mol_column: Adapter.BYTES_TO_MOL}



from hashlib import md5
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import fsspec
from loguru import logger

from polaris.utils.errors import PolarisHubError
from polaris.utils.types import TimeoutTypes
from polaris.utils.misc import sluggify


if TYPE_CHECKING:
    from polaris.hub.client import PolarisHubClient


class PolarisFileSystem(fsspec.AbstractFileSystem):
    """
    A file system interface for accessing datasets on the Polaris platform.

    This class extends `fsspec.AbstractFileSystem` and provides methods to list objects within a Polaris dataset
    and fetch the content of a file from the dataset.

    Note: Zarr Integration
        This file system can be used with Zarr to load multidimensional array data stored in a Dataset from
        the Polaris infrastructure. This class is needed because we otherwise cannot generate signed URLs for
        folders and Zarr is a folder based data-format.

        ```python
        fs = PolarisFileSystem(...)
        store = zarr.storage.FSStore(..., fs=polaris_fs)
        root = zarr.open(store, mode="r")
        ```

    Args:
        polaris_client: The Polaris Hub client used to make API requests.
        dataset_owner: The owner of the dataset.
        dataset_name: The name of the dataset.
    """

    sep = "/"
    protocol = "polarisfs"
    async_impl = False

    def __init__(
        self,
        polaris_client: "PolarisHubClient",
        dataset_owner: str,
        dataset_name: str,
        **kwargs: dict,
    ):
        super().__init__(use_listing_cache=True, listings_expiry_time=None, max_paths=None, **kwargs)

        self.polaris_client = polaris_client
        self.default_timeout = self.polaris_client.settings.default_timeout

        # Prefix to remove from ls entries
        self.prefix = f"dataset/{dataset_owner}/{sluggify(dataset_name)}/"
        self.base_path = f"/storage/{self.prefix.rstrip('/')}"

    @staticmethod
    def is_polarisfs_path(path: str) -> bool:
        """Check if the given path is a PolarisFS path.

        Args:
            path: The path to check.

        Returns:
            True if the path is a PolarisFS path; otherwise, False.
        """
        return path.startswith(f"{PolarisFileSystem.protocol}://")

    def ls(
        self,
        path: str,
        detail: bool = False,
        timeout: Optional[TimeoutTypes] = None,
        **kwargs: dict,
    ) -> Union[List[str], List[Dict[str, Any]]]:
        """List objects in the specified path within the Polaris dataset.

        Args:
            path: The path within the dataset to list objects.
            detail: If True, returns detailed information about each object.
            timeout: Maximum time (in seconds) to wait for the request to complete.

        Returns:
            A list of dictionaries if detail is True; otherwise, a list of object names.
        """
        if timeout is None:
            timeout = self.default_timeout

        ls_path = self.sep.join([self.base_path, "ls", path])

        # GET request to Polaris Hub to list objects in path
        response = self.polaris_client.get(ls_path.rstrip("/"), timeout=timeout)
        response.raise_for_status()

        detailed_listings = [
            {
                "name": p["name"].removeprefix(self.prefix),
                "size": p["size"],
                "type": p["type"],
            }
            for p in response.json()
        ]

        self.dircache[path] = detailed_listings

        if not detail:
            return [p["name"] for p in detailed_listings]

        return detailed_listings

    def cat_file(
        self,
        path: str,
        start: Union[int, None] = None,
        end: Union[int, None] = None,
        timeout: Optional[TimeoutTypes] = None,
        **kwargs: dict,
    ) -> bytes:
        """Fetches and returns the content of a file from the Polaris dataset.

        Args:
            path: The path to the file within the dataset.
            start: The starting index of the content to retrieve.
            end: The ending index of the content to retrieve.
            timeout: Maximum time (in seconds) to wait for the request to complete.
            kwargs: Extra arguments passed to `fsspec.open()`

        Returns:
            The content of the requested file.
        """
        if timeout is None:
            timeout = self.default_timeout

        cat_path = self.sep.join([self.base_path, path])

        # GET request to Polaris Hub for signed URL of file
        response = self.polaris_client.get(cat_path)

        # This should be a 307 redirect with the signed URL
        if response.status_code != 307:
            raise PolarisHubError("Could not get signed URL from Polaris Hub.")

        hub_response_body = response.json()
        signed_url = hub_response_body["url"]

        headers = {"Content-Type": "application/octet-stream", **hub_response_body["headers"]}

        response = self.polaris_client.request(
            url=signed_url,
            method="GET",
            auth=None,
            headers=headers,
            timeout=timeout,
        )
        response.raise_for_status()
        response_content = response.content

        # Verify the checksum on download
        expected_md5sum = self.polaris_client.get_metadata_from_response(response, "md5sum")
        if expected_md5sum is None:
            raise PolarisHubError("MD5 checksum not found in response headers.")
        logger.debug(f"MD5 checksum found in response headers: {expected_md5sum}.")

        md5sum = md5(response_content).hexdigest()
        logger.debug(f"MD5 checksum computed for response content: {md5sum}.")

        if md5sum != expected_md5sum:
            raise PolarisHubError(
                f"MD5 checksum verification failed. Expected {expected_md5sum}, got {md5sum}."
            )

        return response_content[start:end]

    def rm(self, path: str, recursive: bool = False, maxdepth: Optional[int] = None) -> None:
        """Remove a file or directory from the Polaris dataset.

        This method is provided for compatibility with the Zarr storage interface.
        It may be called by the Zarr store when removing a file or directory.

        Args:
            path: The path to the file or directory to be removed.
            recursive: If True, remove directories and their contents recursively.
            maxdepth: The maximum depth to recurse when removing directories.

        Returns:
            None

        Note:
            This method currently it does not perform any removal operations and is included
            as a placeholder that aligns with the Zarr interface's expectations.
        """
        raise NotImplementedError("PolarisFS does not currently support the file removal operation.")

    def pipe_file(
        self,
        path: str,
        content: Union[bytes, str],
        timeout: Optional[TimeoutTypes] = None,
        **kwargs: dict,
    ) -> None:
        """Pipes the content of a file to the Polaris dataset.

        Args:
            path: The path to the file within the dataset.
            content: The content to be piped into the file.
            timeout: Maximum time (in seconds) to wait for the request to complete.

        Returns:
            None
        """
        if timeout is None:
            timeout = self.default_timeout

        pipe_path = self.sep.join([self.base_path, path])

        # PUT request to Polaris Hub to put object in path
        response = self.polaris_client.put(pipe_path, timeout=timeout)

        if response.status_code != 307:
            raise PolarisHubError("Could not get signed URL from Polaris Hub.")

        hub_response_body = response.json()
        signed_url = hub_response_body["url"]

        headers = {"Content-Type": "application/octet-stream", **hub_response_body["headers"]}

        response = self.polaris_client.request(
            url=signed_url,
            method="PUT",
            auth=None,
            content=content,
            headers=headers,
            timeout=timeout,
        )
        response.raise_for_status()

        new_listing = [{"name": path, "type": "file", "size": len(content)}]
        try:
            current_listings = self._ls_from_cache(self._parent(path))
            new_listing.extend(current_listings)
        except FileNotFoundError:
            # self._ls_from_cache() raises FileNotFoundError when no listings
            # are found within a path. We can then assume it's a new directory
            # and new_listing can be used as is.
            pass

        self.dircache[self._parent(path)] = new_listing



import json
from pathlib import Path

from authlib.integrations.httpx_client import OAuth2Auth

from polaris.utils.constants import DEFAULT_CACHE_DIR


class CachedTokenAuth(OAuth2Auth):
    """
    A combination of an authlib token and a httpx auth class, that will cache the token to a file.
    """

    def __init__(
        self,
        token: dict | None = None,
        token_placement="header",
        client=None,
        cache_dir=DEFAULT_CACHE_DIR,
        filename="hub_auth_token.json",
    ):
        self.token_cache_path = Path(cache_dir) / filename

        if token is None and self.token_cache_path.exists():
            token = json.loads(self.token_cache_path.read_text())

        super().__init__(token, token_placement, client)

    def set_token(self, token: dict):
        super().set_token(token)

        # Ensure the cache directory exists.
        self.token_cache_path.parent.mkdir(parents=True, exist_ok=True)

        # We cache afterward, because the token setter adds fields we need to save (i.e. expires_at).
        self.token_cache_path.write_text(json.dumps(token))


class ExternalCachedTokenAuth(CachedTokenAuth):
    """
    Cached token for external authentication.
    """

    def __init__(
        self,
        token: dict | None = None,
        token_placement="header",
        client=None,
        cache_dir=DEFAULT_CACHE_DIR,
        filename="external_auth_token.json",
    ):
        super().__init__(token, token_placement, client, cache_dir, filename)
